"""
A static safe controller for traffic lights using FLOW (https://github.com/flow-project/flow). 

Network:
2 x 2 traffic grid. 
Each edge has two lanes. Vehicles can turn and routes are random. 

"""
import numpy as np
import random
# set random seeds
np.random.seed(5051)
random.seed(6169)

from flow.envs.traffic_light_grid import TrafficLightGridEnv 
from flow.controllers.lane_change_controllers import SimLaneChangeController
from flow.controllers.routing_controllers import GridRouter
from stable_baselines.common.vec_env import DummyVecEnv, SubprocVecEnv
from flow.utils.registry import env_constructor
from flow.core import rewards
from train import run_model_stablebaseline
from flow.core.params import InFlows
from flow.controllers import GridRouter, SimCarFollowingController
from flow.envs import TrafficLightGridPOEnv
from flow.core.experiment import Experiment
from flow.core.params import VehicleParams
from flow.core.params import SumoParams, EnvParams, InitialConfig, NetParams, \
    InFlows, SumoCarFollowingParams, SumoLaneChangeParams
from flow.core.params import TrafficLightParams
from grid_network import TrafficLightGridNetwork
from flow.core.params import NetParams

# configurable parameters
VEHICLE_RATE = 500 # number of vehicles per hour
HORIZON = 3000 # number of training steps per iteration/simulation
RENDER = False

def gen_edges(col_num, row_num):
    """Generate the names of the outer edges in the grid network.
    Parameters
    ----------
    col_num : int
        number of columns in the grid
    row_num : int
        number of rows in the grid
    Returns
    -------
    list of str
        names of all the outer edges
    """
    edges = []

    # build the left and then the right edges
    for i in range(col_num):
        edges += ['left' + str(row_num) + '_' + str(i)]
        edges += ['right' + '0' + '_' + str(i)]

    # build the bottom and then top edges
    for i in range(row_num):
        edges += ['bot' + str(i) + '_' + '0']
        edges += ['top' + str(i) + '_' + str(col_num)]

    return edges


def get_outside_edges(rows, cols):
    edges = []
    for i in range(rows):
        edges.append("bot{}_0".format(i))
        edges.append("top{}_{}".format(i, cols))
    for j in range(cols):
        edges.append("left{}_{}".format(rows, j))
        edges.append("right0_{}".format(j))
    return edges

# grid parameters
inner_length = 100
long_length = 100
short_length = 100
n = 2  # rows
m = 2  # columns
num_cars_left = 10
num_cars_right = 10
num_cars_top = 10
num_cars_bot = 10
tot_cars = (num_cars_left + num_cars_right) * m \
    + (num_cars_top + num_cars_bot) * n

grid_array = {"short_length": short_length, "inner_length": inner_length,
              "long_length": long_length, "row_num": n, "col_num": m,
              "cars_left": num_cars_left, "cars_right": num_cars_right,
              "cars_top": num_cars_top, "cars_bot": num_cars_bot}

# traffic light logic
tl_logic = TrafficLightParams(baseline=False)
"""
Traffic light phases
SUMO traffic lights documentation: https://sumo.dlr.de/docs/Simulation/Traffic_Lights.html#Signal_state_definitions
"""
nodes = ["center0", "center1", "center2", "center3"]
# phases = [{"duration": "31", "state": "GGGrrrGGGrrr"},
#           {"duration": "6", "state": "yyyrrryyyrrr"},
#           {"duration": "31", "state": "rrrGGGrrrGGG"},
#           {"duration": "6", "state": "rrryyyrrryyy"}]
phases = ["ggggrrrrggggrrrr", "yyyyrrrryyyyrrrr", "rrrrggggrrrrgggg", "rrrryyyyrrrryyyy"]
phases1lane = ['GGGrrrGGGrrr', 'yyyrrryyyrrr', 'rrrGGGrrrGGG', 'rrryyyrrryyy']
phases3lanes = ["GGGrrrrrGGGrrrrr", "yyyrrrrryyyrrrrr", "rrrGrrrrrrrGrrrr", "rrryrrrrrrryrrrr", "rrrrGGGrrrrrGGGr", "rrrryyyrrrrryyyr", "rrrrrrrGrrrrrrrG", "rrrrrrryrrrrrrry"]
phases2lanes = ["GGrrrrGGrrrr", "yyrrrryyrrrr", "rrGrrrrrGrrr", "rryrrrrryrrr", "rrrGGrrrrGGr", "rrryyrrrryyr", "rrrrrGrrrrrG", "rrrrryrrrrry"]

durations = [{"duration": "31", "state": "GGGrrrGGGrrr"},
             {"duration": "6", "state": "yyyrrryyyrrr"},
             {"duration": "31", "state": "rrrGGGrrrGGG"},
             {"duration": "6", "state": "rrryyyrrryyy"}]
durations2lanes = [{"duration": "31", "state": "ggggrrrrggggrrrr"},
                   {"duration": "6", "state": "yyyyrrrryyyyrrrr"},
                   {"duration": "31", "state": "rrrrggggrrrrgggg"},
                   {"duration": "6", "state": "rrrryyyyrrrryyyy"}]
durations2lanes_alt = [{"duration": "5", "state": "ggggrrrrggggrrrr"},
                   {"duration": "3", "state": "yyyyrrrryyyyrrrr"},
                   {"duration": "5", "state": "rrrrggggrrrrgggg"},
                   {"duration": "3", "state": "rrrryyyyrrrryyyy"}]
durations3lanes = [{"duration": "12", "state": "GGGrrrrrGGGrrrrr"},
                   {"duration": "3", "state": "yyyrrrrryyyrrrrr"},
                   {"duration": "5", "state": "rrrGrrrrrrrGrrrr"},
                   {"duration": "3", "state": "rrryrrrrrrryrrrr"},
                   {"duration": "12", "state": "rrrrGGGrrrrrGGGr"},
                   {"duration": "3", "state": "rrrryyyrrrrryyyr"},
                   {"duration": "5", "state": "rrrrrrrGrrrrrrrG"},
                   {"duration": "3", "state": "rrrrrrryrrrrrrry"}]
durations2lanes_left_turning = [{"duration": "18", "state": "GGrrrrGGrrrr"},
                   {"duration": "3", "state": "yyrrrryyrrrr"},
                   {"duration": "5", "state": "rrGrrrrrGrrr"},
                   {"duration": "3", "state": "rryrrrrryrrr"},
                   {"duration": "18", "state": "rrrGGrrrrGGr"},
                   {"duration": "3", "state": "rrryyrrrryyr"},
                   {"duration": "5", "state": "rrrrrGrrrrrG"},
                   {"duration": "3", "state": "rrrrryrrrrry"}]

# initialize nodes (intersections)
for node_id in nodes:
    # tl_logic.add(node_id, tls_type="static", programID="1", offset=None)
    tl_logic.add(node_id, tls_type="static", programID="1", offset=None, phases=durations2lanes_left_turning)

# vehicle parameters
vehicles = VehicleParams()
vehicles.add("human",
             acceleration_controller=(SimCarFollowingController, {}),
             lane_change_controller=(SimLaneChangeController, {}),
             lane_change_params=SumoLaneChangeParams(lane_change_mode='strategic'),
             car_following_params=SumoCarFollowingParams(
                 speed_mode="right_of_way",
                 impatience="off", 
                 decel=7.5, 
                #  accel=4.0, 
                 min_gap=2.5))

initial_config = InitialConfig(
    spacing='custom')

# add vehicles to the network
inflows = InFlows()
p = .15
for edge in get_outside_edges(n, m):
    if edge[0] == 'l' or edge[0] == 'r':
        inflows.add(veh_type="human",
                    edge=edge,
                    vehs_per_hour=VEHICLE_RATE,
                    # probability=p,
                    # depart_lane="random",
                    depart_lane="best",
                    depart_speed="random",
                    color="white")
    else:
        inflows.add(veh_type="human",
                    edge=edge,
                    vehs_per_hour=VEHICLE_RATE,
                    # probability=p,
                    depart_lane="best",
                    depart_speed="random",
                    color="white")

# additional parameters
ADDITIONAL_NET_PARAMS = {"grid_array": grid_array, "speed_limit": 35,
                         "horizontal_lanes": 2, "vertical_lanes": 2,
                         "traffic_lights": True}
ADDITIONAL_ENV_PARAMS = {'switch_time': 2.0, 'switch_time_rg': 3.0,'tl_type': 'controlled',
                         'discrete': False, 'num_observed': 6, 'target_velocity': 50, 'phase': phases2lanes}

# network, environment, and simulation parameters
net_params = NetParams(inflows=inflows, additional_params=ADDITIONAL_NET_PARAMS)
env_params = EnvParams(additional_params=ADDITIONAL_ENV_PARAMS)
sim_params = SumoParams(sim_step=0.1, render=RENDER, restart_instance=True)

# flow parameters
flow_params = dict(
    exp_tag='RL_lights',
    env_name=TrafficLightGridPOEnv,
    network=TrafficLightGridNetwork,
    simulator='traci',
    sim=sim_params,
    env=env_params,
    net=net_params,
    veh=vehicles,
    initial=initial_config,
    tls=tl_logic
)

# number of time steps
flow_params['env'].horizon = HORIZON

# model = run_model_stablebaseline(flow_params, rollout_size=60, num_steps=1000)
# #print(dir(model))

# flow_params['sim'].render = True
# env = env_constructor(params=flow_params, version=0)()
# # The algorithms require a vectorized environment to run
# eval_env = DummyVecEnv([lambda: env])
# obs = eval_env.reset()
# reward = 0
# for _ in range(flow_params['env'].horizon):
#     action, _states = model.predict(obs)
#     obs, rewards, dones, info = eval_env.step(action)
#     reward += rewards
# print('the final reward is {}'.format(reward))

exp = Experiment(flow_params)
# to see possible routes
# print(exp.env.network.specify_routes(net_params))

# run the sumo simulation
_ = exp.run(1, convert_to_csv=False)
#print("average delay is:", rewards.min_delay_unscaled(exp.env))
