"""Environments for networks with traffic lights.

These environments are used to train traffic lights to regulate traffic flow
through an n x m traffic light grid.
"""

import numpy as np
import re

from gym.spaces.box import Box
from gym.spaces.discrete import Discrete
from gym.spaces import Tuple

from flow.core import rewards
from .base import Env

ADDITIONAL_ENV_PARAMS = {
    # minimum switch time for each traffic light (in seconds)
    "switch_time": 2.0,
    "switch_time_rg": 3.0,
    # whether the traffic lights should be actuated by sumo or RL
    # options are "controlled" and "actuated"
    "tl_type": "controlled",
    # determines whether the action space is meant to be discrete or continuous
    "discrete": False,
}

ADDITIONAL_PO_ENV_PARAMS = {
    # num of vehicles the agent can observe on each incoming edge
    "num_observed": 2,
    # velocity to use in reward functions
    "target_velocity": 30,
}


class TrafficLightGridEnv(Env):
    """Environment used to train traffic lights.

    Required from env_params:

    * switch_time: minimum time a light must be constant before
      it switches (in seconds).
      Earlier RL commands are ignored.
    * tl_type: whether the traffic lights should be actuated by sumo or RL,
      options are respectively "actuated" and "controlled"
    * discrete: determines whether the action space is meant to be discrete or
      continuous

    States
        An observation is the distance of each vehicle to its intersection, a
        number uniquely identifying which edge the vehicle is on, and the speed
        of the vehicle.

    Actions
        The action space consist of a list of float variables ranging from 0-1
        specifying whether a traffic light is supposed to switch or not. The
        actions are sent to the traffic light in the grid from left to right
        and then top to bottom.

    Rewards
        The reward is the negative per vehicle delay minus a penalty for
        switching traffic lights

    Termination
        A rollout is terminated once the time horizon is reached.

    Additional
        Vehicles are rerouted to the start of their original routes once they
        reach the end of the network in order to ensure a constant number of
        vehicles.

    Attributes
    ----------
    grid_array : dict
        Array containing information on the traffic light grid, such as the
        length of roads, row_num, col_num, number of initial cars
    rows : int
        Number of rows in this traffic light grid network
    cols : int
        Number of columns in this traffic light grid network
    num_traffic_lights : int
        Number of intersection in this traffic light grid network
    tl_type : str
        Type of traffic lights, either 'actuated' or 'static'
    steps : int
        Horizon of this experiment, see EnvParams.horion
    obs_var_labels : dict
        Referenced in the visualizer. Tells the visualizer which
        metrics to track
    node_mapping : dict
        Dictionary mapping intersections / nodes (nomenclature is used
        interchangeably here) to the edges that are leading to said
        intersection / node
    last_change : np array [num_traffic_lights]x1 np array
        Multi-dimensional array keeping track, in timesteps, of how much time
        has passed since the last change to yellow for each traffic light
    direction : np array [num_traffic_lights]x1 np array
        Multi-dimensional array keeping track of which direction in traffic
        light is flowing. 0 indicates flow from top to bottom, and
        1 indicates flow from left to right
    currently_yellow : np array [num_traffic_lights]x1 np array
        Multi-dimensional array keeping track of whether or not each traffic
        light is currently yellow. 1 if yellow, 0 if not
    min_switch_time : np array [num_traffic_lights]x1 np array
        The minimum time in timesteps that a light can be yellow. Serves
        as a lower bound
    discrete : bool
        Indicates whether or not the action space is discrete. See below for
        more information:
        https://github.com/openai/gym/blob/master/gym/spaces/discrete.py
    """

    def __init__(self, env_params, sim_params, network, simulator='traci'):
        self.env_params = env_params

        for p in ADDITIONAL_ENV_PARAMS.keys():
            if p not in env_params.additional_params:
                raise KeyError(
                    'Environment parameter "{}" not supplied'.format(p))

        self.grid_array = network.net_params.additional_params["grid_array"]
        self.rows = self.grid_array["row_num"]
        self.cols = self.grid_array["col_num"]
        # self.num_observed = self.grid_array.get("num_observed", 3)
        self.num_traffic_lights = self.rows * self.cols
        self.tl_type = env_params.additional_params.get('tl_type')

        super().__init__(env_params, sim_params, network, simulator)

        # Saving env variables for plotting
        self.steps = env_params.horizon
        self.obs_var_labels = {
            'edges': np.zeros((self.steps, self.k.vehicle.num_vehicles)),
            'velocities': np.zeros((self.steps, self.k.vehicle.num_vehicles)),
            'positions': np.zeros((self.steps, self.k.vehicle.num_vehicles))
        }

        # Keeps track of the last time the traffic lights in an intersection
        # were allowed to change (the last time the lights were allowed to
        # change from a red-green state to a red-yellow state.)
        # self.last_change = np.zeros((self.rows * self.cols, 1))
        self.last_change = [0 for i in range(self.rows * self.cols)]
        # Keeps track of the direction of the intersection (the direction that
        # is currently being allowed to flow. 0 indicates flow from top to
        # bottom, and 1 indicates flow from left to right.)
        # self.direction = np.zeros((self.rows * self.cols, 1))

        # Keeps track of whether cars are turing left in an intersection
        # 0 indicates cars are not turning left, 1 indicates cars are turning left
        # self.direction_left = np.zeros((self.rows * self.cols, 1))

        # Value of 1 indicates that the intersection is in a red-yellow state.
        # value 0 indicates that the intersection is in a red-green state.
        # self.currently_yellow = np.zeros((self.rows * self.cols, 1))
        self.currently_yellow = [0 for i in range(self.rows * self.cols)]

        # when this hits min_switch_time we change from yellow to red
        # the second column indicates the direction that is currently being
        # allowed to flow. 0 is flowing top to bottom, 1 is left to right
        # For third column, 0 signifies yellow and 1 green or red
        self.min_switch_time = env_params.additional_params["switch_time"]

        # min_switch_time for red and green lights
        self.min_switch_time_rg = env_params.additional_params["switch_time_rg"]

        if self.tl_type != "actuated":
            for i in range(self.rows * self.cols):
                self.k.traffic_light.set_state(
                    node_id='center' + str(i), state=env_params.additional_params['phase'][0])
                self.currently_yellow[i] = 0

        # # Additional Information for Plotting
        # self.edge_mapping = {"top": [], "bot": [], "right": [], "left": []}
        # for i, veh_id in enumerate(self.k.vehicle.get_ids()):
        #     edge = self.k.vehicle.get_edge(veh_id)
        #     for key in self.edge_mapping:
        #         if key in edge:
        #             self.edge_mapping[key].append(i)
        #             break

        # check whether the action space is meant to be discrete or continuous
        self.discrete = env_params.additional_params.get("discrete", False)

        self.num_waiting_previously = [0 for i in range(self.rows * self.cols)]

        self.current_phase = [0 for i in range(self.rows * self.cols)]

        self.num_lanes = self.k.network.num_lanes(self.k.network.get_edge_list()[0])
        self.node_format = re.compile("([a-zA-Z]+)([0-9]+)")
        self.node_mapping_dict = dict(self.network.node_mapping)
        self.intersection_ids = self.node_mapping_dict.keys()

    @property
    def action_space(self):
        """See class definition."""
        if self.discrete:
            return Discrete(4)
        else:
            return Box(
                low=-1,
                high=1,
                shape=(1,),
                dtype=np.float32)

    def _apply_rl_actions(self, rl_actions):
        """See class definition."""
        # print("rl_action is ", rl_actions)

        node, edges = self.network.node_mapping[self.mini_step_counter]
        node_format = re.compile("([a-zA-Z]+)([0-9]+)")
        _, node_idx = node_format.match(node).groups() 
        i = int(node_idx)

        # print("last change ", self.last_change[i])
        # print("currently yellow", self.currently_yellow[i])
        # print("self.min_switch_time_rg: ", self.min_switch_time_rg)
        # print("self.min_switch_time: ", self.min_switch_time)
        self.last_change[i] += 0.1

        # if self.currently_yellow[i] == 1: 
        #     if self.last_change[i] >= self.min_switch_time:
        #         self.k.traffic_light.set_state(
        #             node_id='center{}'.format(i),
        #             state=self.env_params.additional_params['phase'][self.current_phase[i] * 2])
        #         self.last_change[i] = 0.0
        #         self.currently_yellow[i] = 0
        # else: 
        #     phase_change = (rl_actions != self.current_phase[i])
        #     if phase_change and self.last_change[i] >= self.min_switch_time_rg:
        #         self.current_phase[i] = rl_actions
        #         self.k.traffic_light.set_state(
        #             node_id='center{}'.format(i),
        #             state=self.env_params.additional_params['phase'][(self.current_phase[i] * 2 - 1) % 8])
        #         self.last_change[i] = 0.0
        #         self.currently_yellow[i] = 1

        phase_change = (rl_actions != self.current_phase[i])
        if phase_change and self.last_change[i] >= self.min_switch_time_rg:
            self.current_phase[i] = rl_actions
            self.k.traffic_light.set_state(
                node_id='center{}'.format(i),
                state=self.env_params.additional_params['phase'][self.current_phase[i] * 2])
            self.last_change[i] = 0.0


    def compute_reward(self, rl_actions, **kwargs):
        """See class definition."""
        print("reward 1 is called")
        return - rewards.min_delay_unscaled(self) \
            - rewards.boolean_action_penalty(rl_actions >= 0.5, gain=1.0)

    # ===============================
    # ============ UTILS ============
    # ===============================

    def get_distance_to_intersection(self, veh_ids):
        """Determine the distance from a vehicle to its next intersection.

        Parameters
        ----------
        veh_ids : str or str list
            vehicle(s) identifier(s)

        Returns
        -------
        float (or float list)
            distance to closest intersection
        """
        if isinstance(veh_ids, list):
            return [self.get_distance_to_intersection(veh_id)
                    for veh_id in veh_ids]
        return self.find_intersection_dist(veh_ids)

    def find_intersection_dist(self, veh_id):
        """Return distance from intersection.

        Return the distance from the vehicle's current position to the position
        of the node it is heading toward.
        """
        edge_id = self.k.vehicle.get_edge(veh_id)
        # FIXME this might not be the best way of handling this
        if edge_id == "":
            return -10
        if 'center' in edge_id:
            return 0
        edge_len = self.k.network.edge_length(edge_id)
        relative_pos = self.k.vehicle.get_position(veh_id)
        dist = edge_len - relative_pos
        return dist

    def _convert_edge(self, edges):
        """Convert the string edge to a number.

        Start at the bottom left vertical edge and going right and then up, so
        the bottom left vertical edge is zero, the right edge beside it  is 1.

        The numbers are assigned along the lowest column, then the lowest row,
        then the second lowest column, etc. Left goes before right, top goes
        before bottom.

        The values are zero indexed.

        Parameters
        ----------
        edges : list of str or str
            name of the edge(s)

        Returns
        -------
        list of int or int
            a number uniquely identifying each edge
        """
        if isinstance(edges, list):
            return [self._split_edge(edge) for edge in edges]
        else:
            return self._split_edge(edges)

    def _split_edge(self, edge):
        """Act as utility function for convert_edge."""
        if edge:
            if edge[0] == ":":  # center
                center_index = int(edge.split("center")[1][0])
                base = ((self.cols + 1) * self.rows * 2) \
                    + ((self.rows + 1) * self.cols * 2)
                return base + center_index + 1
            else:
                pattern = re.compile(r"[a-zA-Z]+")
                edge_type = pattern.match(edge).group()
                edge = edge.split(edge_type)[1].split('_')
                row_index, col_index = [int(x) for x in edge]
                if edge_type in ['bot', 'top']:
                    rows_below = 2 * (self.cols + 1) * row_index
                    cols_below = 2 * (self.cols * (row_index + 1))
                    edge_num = rows_below + cols_below + 2 * col_index + 1
                    return edge_num if edge_type == 'bot' else edge_num + 1
                if edge_type in ['left', 'right']:
                    rows_below = 2 * (self.cols + 1) * row_index
                    cols_below = 2 * (self.cols * row_index)
                    edge_num = rows_below + cols_below + 2 * col_index + 1
                    return edge_num if edge_type == 'left' else edge_num + 1
        else:
            return 0

    def _get_relative_node(self, agent_id, direction):
        """Yield node number of traffic light agent in a given direction.

        For example, the nodes in a traffic light grid with 2 rows and 3
        columns are indexed as follows:

            |     |     |
        --- 3 --- 4 --- 5 ---
            |     |     |
        --- 0 --- 1 --- 2 ---
            |     |     |

        See flow.networks.traffic_light_grid for more information.

        Example of function usage:
        - Seeking the "top" direction to ":center0" would return 3.
        - Seeking the "bottom" direction to ":center0" would return -1.

        Parameters
        ----------
        agent_id : str
            agent id of the form ":center#"
        direction : str
            top, bottom, left, right

        Returns
        -------
        int
            node number
        """
        ID_IDX = 1
        agent_id_num = int(agent_id.split("center")[ID_IDX])
        if direction == "top":
            node = agent_id_num + self.cols
            if node >= self.cols * self.rows:
                node = -1
        elif direction == "bottom":
            node = agent_id_num - self.cols
            if node < 0:
                node = -1
        elif direction == "left":
            if agent_id_num % self.cols == 0:
                node = -1
            else:
                node = agent_id_num - 1
        elif direction == "right":
            if agent_id_num % self.cols == self.cols - 1:
                node = -1
            else:
                node = agent_id_num + 1
        else:
            raise NotImplementedError

        return node

    def additional_command(self):
        """See parent class.

        Used to insert vehicles that are on the exit edge and place them
        back on their entrance edge.
        """
        for veh_id in self.k.vehicle.get_ids():
            self._reroute_if_final_edge(veh_id)

    def _reroute_if_final_edge(self, veh_id):
        """Reroute vehicle associated with veh_id.

        Checks if an edge is the final edge. If it is return the route it
        should start off at.
        """
        edge = self.k.vehicle.get_edge(veh_id)
        if edge == "":
            return
        if edge[0] == ":":  # center edge
            return
        pattern = re.compile(r"[a-zA-Z]+")
        edge_type = pattern.match(edge).group()
        edge = edge.split(edge_type)[1].split('_')
        row_index, col_index = [int(x) for x in edge]

        # find the route that we're going to place the vehicle on if we are
        # going to remove it
        route_id = None
        if edge_type == 'bot' and col_index == self.cols:
            route_id = "bot{}_0".format(row_index)
        elif edge_type == 'top' and col_index == 0:
            route_id = "top{}_{}".format(row_index, self.cols)
        elif edge_type == 'left' and row_index == 0:
            route_id = "left{}_{}".format(self.rows, col_index)
        elif edge_type == 'right' and row_index == self.rows:
            route_id = "right0_{}".format(col_index)

        if route_id is not None:
            type_id = self.k.vehicle.get_type(veh_id)
            lane_index = self.k.vehicle.get_lane(veh_id)
            # remove the vehicle
            self.k.vehicle.remove(veh_id)
            # reintroduce it at the start of the network
            self.k.vehicle.add(
                veh_id=veh_id,
                edge=route_id,
                type_id=str(type_id),
                lane=str(lane_index),
                pos="0",
                speed="max")

    def get_closest_to_intersection(self, edges, num_closest, padding=False):
        """Return the IDs of the vehicles that are closest to an intersection 
        given a maximum number of observed vehicles on each lane. The number 
        is defined as num_closest / number of lanes

        For each edge in edges, return the IDs (veh_id) of the num_closest
        vehicles in edge that are closest to an intersection (the intersection
        they are heading towards).

        This function performs no check on whether or not edges are going
        towards an intersection or not, it just gets the vehicles that are
        closest to the end of their edges.

        If there are less than num_closest vehicles on an edge, the function
        performs padding by adding empty strings "" instead of vehicle ids if
        the padding parameter is set to True.

        Parameters
        ----------
        edges : str | str list
            ID of an edge or list of edge IDs.
        num_closest : int (> 0)
            Number of vehicles to consider on each edge.
        padding : bool (default False)
            If there are less than num_closest vehicles on an edge, perform
            padding by adding empty strings "" instead of vehicle ids if the
            padding parameter is set to True (note: leaving padding to False
            while passing a list of several edges as parameter can lead to
            information loss since you will not know which edge, if any,
            contains less than num_closest vehicles).

        Usage
        -----
        For example, consider the following network, composed of 4 edges
        whose ids are "edge0", "edge1", "edge2" and "edge3", the numbers
        being vehicles all headed towards intersection x. The ID of the vehicle
        with number n is "veh{n}" (edge "veh0", "veh1"...).

                            edge1
                            |   |
                            | 7 |
                            | 8 |
               -------------|   |-------------
        edge0    1 2 3 4 5 6  x                 edge2
               -------------|   |-------------
                            | 9 |
                            | 10|
                            | 11|
                            edge3

        And consider the following example calls on the previous network:

        >>> get_closest_to_intersection("edge0", 4)
        ["veh6", "veh5", "veh4", "veh3"]

        >>> get_closest_to_intersection("edge0", 8)
        ["veh6", "veh5", "veh4", "veh3", "veh2", "veh1"]

        >>> get_closest_to_intersection("edge0", 8, padding=True)
        ["veh6", "veh5", "veh4", "veh3", "veh2", "veh1", "", ""]

        >>> get_closest_to_intersection(["edge0", "edge1", "edge2", "edge3"],
                                         3, padding=True)
        ["veh6", "veh5", "veh4", "veh8", "veh7", "", "", "", "", "veh9",
         "veh10", "veh11"]

        Returns
        -------
        str list
            If n is the number of edges given as parameters, then the returned
            list contains n * num_closest vehicle IDs.

        Raises
        ------
        ValueError
            if num_closest <= 0
        """
        if num_closest <= 0:
            raise ValueError("Function get_closest_to_intersection called with"
                             "parameter num_closest={}, but num_closest should"
                             "be positive".format(num_closest))

        if isinstance(edges, list):
            ids = [self.get_closest_to_intersection(edge, num_closest)
                   for edge in edges]
            # flatten the list and return it
            return [veh_id for sublist in ids for veh_id in sublist]

        # get the ids of all the vehicles on the edge 'edges' ordered by
        # increasing distance to end of edge (intersection)
        veh_ids_ordered = sorted(self.k.vehicle.get_ids_by_edge(edges),
                                 key=self.get_distance_to_intersection)

        # vehicles closest to an intersection
        closest_veh_ids = veh_ids_ordered[:num_closest]
        # max number of vehicles to observe per lane
        max_num_per_lane = num_closest / self.k.network.num_lanes(edges)
        # count the number of observed vehicles per lane
        vehs_count_per_lane = {} 
        # array of observed vehicle ids on an edge given a maximum number of observations allowed on each lane
        observed_veh_ids = []

        for id in closest_veh_ids: 
            lane = self.k.vehicle.get_lane(id) 
            if lane in vehs_count_per_lane: 
                vehs_count_per_lane[lane] += 1
                # haven't reached the maximum num of observations of this lane
                if vehs_count_per_lane[lane] <= max_num_per_lane:
                    observed_veh_ids.append(id)
            else: 
                vehs_count_per_lane[lane] = 1
                observed_veh_ids.append(id)

        # return the ids of the num_closest vehicles closest to the
        # intersection, potentially with ""-padding.
        pad_lst = [""] * (len(observed_veh_ids) - len(veh_ids_ordered))

        return observed_veh_ids + (pad_lst if padding else [])

    def get_edge_waiting_vehicle_count(self, edge):
        veh_ids = self.k.vehicle.get_ids_by_edge(edge)
        count = 0
        for id in veh_ids:
            if self.k.vehicle.get_speed(id) < 0.1:
                count += 1

        return count

    def get_lane_waiting_vehicle_count(self, edge):
        veh_ids = self.k.vehicle.get_ids_by_edge(edge)
        waiting_count_dict = {}
        for lane_idx in range(self.k.network.num_lanes(edge)): 
            waiting_count_dict[lane_idx] = 0
        for id in veh_ids:
            if self.k.vehicle.get_speed(id) < 0.1:
                waiting_count_dict[self.k.vehicle.get_lane(id)] += 1

        return waiting_count_dict

    def get_info_per_edge(self, edge):
        """
        Returns a list of vehicle info, including speed, distance to intersection, and edge number
        """

        # if isinstance(edges, list):
        #     nums = [self.get_info_per_edge(edge)
        #            for edge in edges]
        #     # flatten the list and return it
        #     return [num_vehs for sublist in nums for num_vehs in sublist]

        # now add in the density and average velocity of the edge
        edge_info = []
        
        lane_ids_dict = {}
        for lane_idx in range(self.k.network.num_lanes(edge)): 
            lane_ids_dict[lane_idx] = []

        ids = self.k.vehicle.get_ids_by_edge(edge)
        for id in ids: 
            lane_ids_dict[self.k.vehicle.get_lane(id)].append(id)

        # print("dict is ", lane_ids_dict)

        waiting_count_dict = self.get_lane_waiting_vehicle_count(edge)
        # print("waiting count dict is ", waiting_count_dict)

        for lane_idx in lane_ids_dict:
            lane_ids = lane_ids_dict[lane_idx]
            if len(lane_ids) > 0:
                vehicle_length = 5
                edge_info += [vehicle_length * len(lane_ids) /
                            self.k.network.edge_length(edge)]
                edge_info += [np.mean(
                    [self.k.vehicle.get_speed(veh_id) for veh_id in
                        lane_ids]) / self.max_speed]
            else:
                edge_info += [0, 1]
            edge_info.append(waiting_count_dict[lane_idx])
        return edge_info

class TrafficLightGridPOEnv(TrafficLightGridEnv):
    """Environment used to train traffic lights.

    Required from env_params:

    * switch_time: minimum switch time for each traffic light (in seconds).
      Earlier RL commands are ignored.
    * num_observed: number of vehicles nearest each intersection that is
      observed in the state space; defaults to 2

    States
        An observation is the number of observed vehicles in each intersection
        closest to the traffic lights, a number uniquely identifying which
        edge the vehicle is on, and the speed of the vehicle.

    Actions
        The action space consist of a list of float variables ranging from 0-1
        specifying whether a traffic light is supposed to switch or not. The
        actions are sent to the traffic light in the grid from left to right
        and then top to bottom.

    Rewards
        The reward is the delay of each vehicle minus a penalty for switching
        traffic lights

    Termination
        A rollout is terminated once the time horizon is reached.

    Additional
        Vehicles are rerouted to the start of their original routes once they
        reach the end of the network in order to ensure a constant number of
        vehicles.

    """

    def __init__(self, env_params, sim_params, network, simulator='traci'):
        super().__init__(env_params, sim_params, network, simulator)

        for p in ADDITIONAL_PO_ENV_PARAMS.keys():
            if p not in env_params.additional_params:
                raise KeyError(
                    'Environment parameter "{}" not supplied'.format(p))

        # number of vehicles nearest each intersection that is observed in the
        # state space; defaults to 2
        self.num_observed = env_params.additional_params.get("num_observed", 2)

        # max speed in the network
        self.max_speed = max(
            self.k.network.speed_limit(edge)
            for edge in self.k.network.get_edge_list())

        # used during visualization
        self.observed_ids = []

    @property
    def observation_space(self):
        """State space that is partially observed.


                                                 +--+--+
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |20|  |
                                +---------------------------------------+
                                |                |     |19              |
                                +----------------+     +----------------+
                                |              17|     |                |
                        +--+-------------------------------------------------+--+
                        |  |  |                  |  |18|                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        | 5|  |                  |16|  |                  |15|  |
        +------------------------------------------------------------------------------------------+
        |               |     |4                 |     |11                |     |14                |
        +---------------+     +------------------+     +------------------+     +------------------+
        |              2|     |                 1|     |                12|     |                  |
        +------------------------------------------------------------------------------------------+
                        |  |3 |                  |  |6 |                  |  |13|
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |  |  |                  |  |  |
                        |  |  |                  |10|  |                  |  |  |
                        +--+-------------------------------------------------+---
                                |                |     |9                 |
                                +----------------+     +------------------+
                                |               7|     |                  |
                                +-------------------+---------------------+
                                                 |  | 8|
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 |  |  |
                                                 +--+---

        A more complicated feature vector
        vector size: 6 * 20 * 3 + 20 + 20 * 2 = 420
        - track [num observed: 6 (default)] vehicles on each edge
        - track [number of edges: 20] edges
        - for each tracked vehicle, record the speed, distance to the intersection, 
          and edge number.
        - track last change, direction, direction_left, and currently yellow for each intersection 
        - track the density and average velocity for each edge.
        

        """
        tl_box = Box(
            low=0.,
            high=1,
            shape=(125,), # 5 intersections in view
            dtype=np.float32)
        return tl_box

    def get_state(self):
        feature_vec = []
        intersection_id = "center{}".format(self.mini_step_counter)
        feature_vec.append(self.get_intersection_state(intersection_id))
        neighbors = self.neighbor_intersections(intersection_id)
        for i, nbr_id in enumerate(neighbors):
            if nbr_id in self.intersection_ids:
                nbr_feature = self.get_intersection_state(nbr_id)
                feature_vec.append(nbr_feature)
            else: 
                reflect_map = [2, 3, 0, 1]
                reflect_to_feature = self.get_intersection_state(neighbors[reflect_map[i]])
                features_per_edge = (reflect_to_feature.size - 1) // 4 # 6 features per edge
                reflected_feature = np.concatenate((reflect_to_feature[features_per_edge * 2 : features_per_edge * 3],
                                                reflect_to_feature[features_per_edge * 3 : features_per_edge * 4],
                                                reflect_to_feature[:features_per_edge],
                                                reflect_to_feature[features_per_edge : features_per_edge * 2],
                                                [reflect_to_feature[-1]]))
                feature_vec.append(reflected_feature)

        return np.concatenate(feature_vec)

    def get_intersection_state(self, intersection_id):
        feature_vec = []
        edges = self.node_mapping_dict[intersection_id]
        # print("mini step counter ", self.mini_step_counter)
        # print("node mapping is ", self.network.node_mapping)
        # print("node is ", node, "edges ", edges)
        idx = self.mini_step_counter
        # print("idx is ", idx)
        for edge in edges:
            feature_vec += self.get_info_per_edge(edge)
        feature_vec.append(self.last_change[idx])
        return np.array(feature_vec)

    def neighbor_intersections(self, intersection_id):
        neighbors = []
        _, node_idx = self.node_format.match(intersection_id).groups()
        idx = int(node_idx)
        # print("idx is ", idx)
        # in the order of LEFT(bot edge), 
        #                 BOTTOM (right edge), 
        #                 RIGHT (top edge), 
        #                 TOP (left edge)
        neighbors.append("center{}".format(idx - 1))
        neighbors.append("center{}".format(idx - self.cols))
        neighbors.append("center{}".format(idx + 1))
        neighbors.append("center{}".format(idx + self.cols))

        return neighbors



    def compute_reward(self, rl_actions, **kwargs):
        """See class definition."""
        idx = self.mini_step_counter
        node, edges = self.network.node_mapping[idx]

        # print("reward node is ", node)
        
        waiting_count = 0
        for edge in edges: 
            waiting_count += self.get_edge_waiting_vehicle_count(edge)
        diff = self.num_waiting_previously[idx] - waiting_count
        self.num_waiting_previously[idx] = waiting_count
        return diff
        
    def additional_command(self):
        """See class definition."""
        # specify observed vehicles
        [self.k.vehicle.set_observed(veh_id) for veh_id in self.observed_ids]


class TrafficLightGridTestEnv(TrafficLightGridEnv):
    """
    Class for use in testing.

    This class overrides RL methods of traffic light grid so we can test
    construction without needing to specify RL methods
    """

    def _apply_rl_actions(self, rl_actions):
        """See class definition."""
        pass

    def compute_reward(self, rl_actions, **kwargs):
        """No return, for testing purposes."""
        return 0
